#include "../object.h"

// Submodules
#include "json/json.h"
#include "json/layout.h"
#include "json-ld/json-ld.h"

// Standard Library
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stddef.h>
#include <limits.h>

extern struct json_field_type ap_signature_composite_type;
extern struct json_field_type ap_activity_source_type;

// TODO: deprecate this in favor of vtable->type_String and factory lookup
// https://www.w3.org/TR/activitystreams-vocabulary/#h-activity-types
struct json_enum ap_object_type_enum[] = {
	{ "Undo", ap_Undo },
	{ "Follow", ap_Follow },
	{ "Delete", ap_Delete },
	{ "Accept", ap_Accept },
	{ "Create", ap_Create },
	{ "TentativeAccept", ap_TentativeAccept },
	{ "Add", ap_Add },
	{ "Arrive", ap_Arrive },
	{ "Ignore", ap_Ignore },
	{ "Join", ap_Join },
	{ "Leave", ap_Leave },
	{ "Like", ap_Like },
	{ "Offer", ap_Offer },
	{ "Invite", ap_Invite },
	{ "Reject", ap_Reject },
	{ "TentativeReject", ap_TentativeReject },
	{ "Remove", ap_Remove },
	{ "Update", ap_Update },
	{ "View", ap_View },
	{ "Listen", ap_Listen },
	{ "Read", ap_Read },
	{ "Move", ap_Move },
	{ "Travel", ap_Travel },
	{ "Announce", ap_Announce },
	{ "Block", ap_Block },
	{ "Flag", ap_Flag },
	{ "Dislike", ap_Dislike },
	{ "EmojiReact", pleroma_EmojiReact },

	{ "Note", ap_Note },
	{ "Question", ap_Question },
	{ "Document", ap_Document },
	{ "Image", ap_Image },
	{ "Tombstone", ap_Tombstone },

	{ "Person", ap_Person },
	{ "Service", ap_Service },
	{ "Application", ap_Application },
	{ "Group", ap_Group },
	{ "Organization", ap_Organization },

	{ "Collection", ap_Collection },
	{ "CollectionPage", ap_CollectionPage },
	{ "OrderedCollection", ap_OrderedCollection },
	{ "OrderedCollectionPage", ap_OrderedCollectionPage },

	// Musicbox Types
	{ "Track", musicbox_Track },
	{ "Album", musicbox_Album },
	{ "Artist", musicbox_Artist },

	// Forge Types

	// Schema Types
	{ "PropertyValue", schema_PropertyValue },

	{ NULL, 0 },
};

struct json_tagged_union_typelist signature_types[] = {
	{
		.tag_value = 1,
		.offset = offsetof( struct ap_object, signature ),
		.type = &ap_signature_composite_type,
	},
	JSON_TAGGED_UNION_TYPELIST_END,
};

extern struct json_field_type endpoints_type;
extern struct json_field_type capabilities_type;
extern struct json_field_type total_items_type;

#define OBJ_TYPE struct ap_public_key
static struct json_object_field ap_public_key_layout[] = {
	JSON_FIELD_STRING( id, true ),
	JSON_FIELD_FIXED_STRING( type, "Key", false ),
	JSON_FIELD_STRING( owner, true ),
	{
		.key = "publicKeyPem",
		.offset = offsetof( OBJ_TYPE, public_key ),
		.required = true,
		.type = &json_field_string,
	},
	JSON_FIELD_END,
};
#undef OBJ_TYPE

void ap_public_key_free( struct ap_public_key* pk )
{
	if( !pk ) { return; }

	free( pk->id );
	free( pk->owner );
	free( pk->public_key );

	free( pk );
}

JSON_FIELD_TYPE_OBJECT_LAYOUT_WITH_DEFAULTS( ap_public_key );

#define OBJ_TYPE struct ap_object
struct json_object_field ap_object_layout[] = {
	// JSON-LD Fields
	// To be compatible with https://w3c.github.io/json-ld-streaming/, @context must come first, then @type (type in ActivityStreams). @id is recommended
	// to be the next field
	{
		.key = "@context",
		.offset = offsetof( OBJ_TYPE, ap_context ),
		.type = &ap_activity_context_type
	},
	JSON_FIELD_ENUM( type, ap_object_type_enum, true ),
	JSON_FIELD_STRING(id,false),

	// Object Fields
	{
		.key = "attachment",
		.offset = offsetof( OBJ_TYPE, attachments ),
		.allow_drop_empty = true,
		.type = &json_field_array_of,
		.array_item_type = &ap_object_type
	},
	{
		.key = "attributedTo",
		.offset = offsetof(OBJ_TYPE,attributed_to),
		.type = &json_field_string,
	},
	{
		.key = "content",
		.offset = offsetof( OBJ_TYPE, content ),
		.type = &ap_activity_source_type
	},
	JSON_FIELD_STRING(name,false),
	{
		.key = "endTime",
		.offset = offsetof( OBJ_TYPE, end_time ),
		.required = false,
		.allow_drop_empty = true,
		.type = &json_field_date_time,
	},
	{
		.key = "inReplyTo",
		.offset = offsetof( OBJ_TYPE, in_reply_to ),
		.type = &json_field_string
	},
	{
		.key = "quoteUrl",
		.offset = offsetof( OBJ_TYPE, quote_url ),
		.type = &json_field_string,
	},
	{
		.key = "published",
		.offset = offsetof( OBJ_TYPE, published ),
		.allow_drop_empty = true,
		.type = &json_field_date_time,
	},
	JSON_FIELD_STRING(summary,false),
	{
		.key = "tag",
		.offset = offsetof( OBJ_TYPE, tags ),
		.type = &json_field_array_of,
		.allow_drop_empty = true,
		.array_item_type = &ap_activity_tag_type
	},
	{
		.key = "updated",
		.offset = offsetof( OBJ_TYPE, updated ),
		.allow_drop_empty = true,
		.type = &json_field_date_time,
	},
	{
		.key = "url",
		.offset = offsetof( OBJ_TYPE, url ),
		.allow_drop_empty = true,
		.type = &json_field_single_or_array_of,
		.array_item_type = &ap_object_ptr_or_ref_type,
	},
	{
		.key = "to",
		.offset = offsetof( OBJ_TYPE, to ),
		.allow_drop_empty = true,
		.type = &json_field_single_or_array_of,
		.array_item_type = &json_field_string,
	},
	{
		.key = "cc",
		.offset = offsetof( OBJ_TYPE, cc ),
		.allow_drop_empty = true,
		.type = &json_field_single_or_array_of,
		.array_item_type = &json_field_string,
	},
	{
		.key = "bcc",
		.offset = offsetof( OBJ_TYPE, bcc ),
		.allow_drop_empty = true,
		.type = &json_field_single_or_array_of,
		.array_item_type = &json_field_string,
	},
	{
		.key = "bto",
		.offset = offsetof( OBJ_TYPE, bto ),
		.allow_drop_empty = true,
		.type = &json_field_single_or_array_of,
		.array_item_type = &json_field_string,
	},
	{
		.key = "mediaType",
		.offset = offsetof( OBJ_TYPE, media_type ),
		.allow_drop_empty = true,
		.type = &json_field_string,
	},

	// PropertyValue-specific fields
	JSON_FIELD_STRING( value, false ),

	// Link-specific fields
	JSON_FIELD_STRING(href,false),
	JSON_FIELD_STRING(rel,false),

	// Activity-specific fields
	JSON_FIELD_STRING(actor,false),
	{
		.key = "object",
		.offset = offsetof( OBJ_TYPE, object ),
		.type = &ap_object_ptr_or_ref_type,
	},
	JSON_FIELD_STRING( target, false ),

	// Actor-specific fields
	JSON_FIELD_STRING(inbox,false),
	JSON_FIELD_STRING(outbox,false),
	{
		.key = "preferredUsername",
		.offset = offsetof( OBJ_TYPE, preferred_username ),
		.type = &json_field_string,
	},
	{
		.key = "endpoints",
		.type = &endpoints_type,
	},
	JSON_FIELD_STRING( featured, false ),
	JSON_FIELD_STRING( followers, false ),
	JSON_FIELD_STRING( following, false ),
	{
		.key = "capabilities",
		.required = false,
		.type = &capabilities_type,
	},
	{
		.key = "icon",
		.offset = offsetof( OBJ_TYPE, icon ),
		.type = &ap_object_type,
	},
	{
		.key = "image",
		.offset = offsetof( OBJ_TYPE, image ),
		.type = &ap_object_type,
	},
	{
		.key = "discoverable",
		.offset = offsetof( OBJ_TYPE, discoverable ),
		.type = &json_field_rich_bool,
	},
	{
		.key = "manuallyApprovesFollowers",
		.offset = offsetof( OBJ_TYPE, manually_approves_followers ),
		.type = &json_field_rich_bool,
	},
	{
		.key = "publicKey",
		.offset = offsetof( OBJ_TYPE, public_key ),
		.type = &ap_public_key_type,
	},
	JSON_FIELD_STRING(context,false),
	{
		.key = "directMessage",
		.offset = offsetof( OBJ_TYPE, direct_message ),
		.type = &json_field_bool
	},
	JSON_FIELD_STRING(conversation,false),
	{
		.key = "source",
		.offset = offsetof( OBJ_TYPE, source ),
		.type = &ap_activity_source_type
	},
	{
		.key = "sensitive",
		.offset = offsetof( OBJ_TYPE, sensitive ),
		.type = &json_field_bool_or_null
	},
	{
		.key = "alsoKnownAs",
		.offset = offsetof( OBJ_TYPE, also_known_as ),
		.allow_drop_empty = true,
		.type = &json_field_array_of,
		.array_item_type = &json_field_string,
	},

	// Collection fields
	{
		.key = "first",
		.offset = offsetof( OBJ_TYPE, first ),
		.type = &ap_object_ptr_or_ref_type,
	},
	{
		.key = "next",
		.offset = offsetof( OBJ_TYPE, next ),
		.type = &ap_object_ptr_or_ref_type,
	},
	JSON_FIELD_STRING( prev, false ),
	{
		.key = "partOf",
		.offset = offsetof( OBJ_TYPE, part_of ),
		.type = &json_field_string,
	},
	{
		.key = "totalItems",
		.offset = 0, //offsetof( struct ap_object, total_items ),
		.type = &total_items_type,
	},
	{
		.key = "items",
		.offset = offsetof( OBJ_TYPE, collection_items ),
		.allow_drop_empty = true,
		.type = &json_field_array_of,
		.array_item_type = &ap_object_ptr_or_ref_type,
	},
	{
		.key = "orderedItems",
		.offset = offsetof( OBJ_TYPE, ordered_items ),
		.allow_drop_empty = true,
		.type = &json_field_array_of,
		.array_item_type = &ap_object_ptr_or_ref_type,
	},

	// Question fields
	{
		.key = "oneOf",
		.offset = offsetof( OBJ_TYPE, one_of ),
		.allow_drop_empty = true,
		.type = &json_field_array_of,
		.array_item_type = &ap_object_ptr_or_ref_type,
	},
	{
		.key = "anyOf",
		.offset = offsetof( OBJ_TYPE, any_of ),
		.allow_drop_empty = true,
		.type = &json_field_array_of,
		.array_item_type = &ap_object_ptr_or_ref_type,
	},
	{
		.key = "closed",
		.offset = offsetof( OBJ_TYPE, closed ),
		.allow_drop_empty = true,
		.type = &json_field_date_time,
	},

	// Other
	JSON_FIELD_STRING(state,false),
	{
		.key = "replies",
		.offset = offsetof( OBJ_TYPE, replies ),
		.type = &ap_object_ptr_or_ref_type,
	},

	// Signature
	{
		.key = "signature",
		.offset = offsetof( OBJ_TYPE, has_signature ),
		.type = &json_field_tagged_union,
		.tagged_item_types = signature_types
	},

	// Extra field storage
	{
		.key = "*",
		.offset = offsetof( OBJ_TYPE, extra ),
		.type = &json_field_extra
	},

	JSON_FIELD_END,
};
#undef OBJ_TYPE
static struct json_layout ap_object_layout_new = {
	.stride = sizeof(struct json_object_field),
	.count = sizeof(ap_object_layout)/sizeof(ap_object_layout[0]) - 1,
	.first_offset = &ap_object_layout[0].offset,
	.first_reflection = (struct json_reflection*)&ap_object_layout[0].key,
};

void* ap_object_alloc_shim() {
	struct ap_object* a;
	a = malloc(sizeof(*a));
	memset(a,0,sizeof(*a));
	return a;
}
void ap_object_free_shim( void* ptr )
{
	ap_object_free(ptr);
}
bool ap_object_writer( struct json_writer* jw, const char* field_name, void* field_data, struct json_reflection* json_field_data, int offset )
{
	struct ap_object* obj = *(void**)field_data;

	if( !json_field_data->required && !obj ) {
		return false;
	}

	json_write_field_name(jw,field_name);
	if( !obj ) {
		fprintf( jw->f, "null" );
		return true;
	}

	if( obj->vtable && obj->vtable->layout ) {
		json_write_pretty_layout( jw, obj->vtable->layout, obj );
	} else {
		json_write_pretty_object_layout( jw, ap_object_layout, obj );
	}

	return true;
}
extern struct ap_object_vtable* factories;

static bool ap_object_downcast( struct json_value buffer, void** data, struct json_layout** new_layout )
{
	struct json_value type;
	struct json_value context;
	bool have_context = false;

	if( !json_value_object_get(buffer,"type",&type) || type.type != jvt_string ) {
		return false;
	}
	if( json_value_object_get(buffer,"@context",&context) ) {
		have_context = true;
	}

	char* json_ld_type = json_ld_lookup_value( context, type.s );

	//printf( "ap_object type=%s (%s)\n", type.s, json_ld_type );

	struct ap_object_vtable* best = NULL;

	// Try by JSON-LD value first
	if( have_context && json_ld_type ) {
		for( struct ap_object_vtable* iter = factories; iter; iter = iter->next ) {
			//printf( "factory type: %s (%d)\n", iter->type_string, iter->type );
			if( !iter->json_ld_type ) { continue; }

			if( 0 == strcmp(json_ld_type,iter->json_ld_type) ) {
				best = iter;
			}
		}
	}
	free(json_ld_type);

	// If we still don't have a type, match by type_string
	if( !best ) {
		for( struct ap_object_vtable* iter = factories; iter; iter = iter->next ) {
			//printf( "factory type: %s (%d)\n", iter->type_string, iter->type );
			if( 0 == strcmp(type.s,iter->type_string) ) {
				best = iter;
			}
		}
	}

	if( best ) {
		printf( "Using type %s (%s)\n", best->type_string, best->json_ld_type );
		struct ap_object* act = *data = best->alloc();
		act->vtable = best;
		*new_layout = best->layout;

		return true;
	}

	//printf( "using default ap_object\n" );

	// Use default
	struct ap_object* act = ap_object_new();
	*data = act;
	*new_layout = &ap_object_layout_new;

	return true;
}
static bool ap_object_reader( struct json_pull_parser* jpp, void* field_data, struct json_reflection* json_field_data, int offset )
{
	void* result = NULL;
	if( !json_read_object_layout_with_downcast( jpp, ap_object_layout, &result, ap_object_downcast ) ) {
		ap_object_free(result);
		printf( "Failed to read object\n" );
		return false;
	}

	*(void**)field_data = result;
	return true;
}
struct json_field_type ap_object_type = {
	.reader = ap_object_reader,  //json_field_object_type_reader,
	.writer = ap_object_writer,
	.size = sizeof( struct name* ),
	.layout = ap_object_layout,
	.alloc = ap_object_alloc_shim,
	.free = ap_object_free_shim,
};

struct ap_object* ap_object_from_FILE( FILE* f )
{
	void* data = NULL;
	if( !json_read_object_layout_from_FILE_ex( f, ap_object_layout, &data, 0, ap_object_downcast ) ) {
		ap_object_free(data);
		return NULL;
	}

	return data;
}
struct ap_object* ap_object_from_file( const char* filename )
{
	FILE* f = fopen(filename,"r");
	if( !f ) { return NULL; }

	struct ap_object* obj = ap_object_from_FILE(f);
	return obj;
}

/*
void ap_object_write_to_FILE( struct ap_object* obj, FILE* f )
{
	struct json_writer jw = {
		.f = f,
		.indentation = "\t",
		.indent = 0,
	};
	if( !obj ) {
		fprintf( f, "null" );
		return;
	}
	json_write_pretty_object_layout( &jw, ap_object_layout, obj );
}
*/

